""" rect_2d.py file
    Rect2D: class name
    Class attributes: _top_left, _size, _is_valid
    TODO: add test for RECT2D
    TODO: add eq hash neq
    The model and naming rules are depend on soccer simulator environment
              -34.0
                |
                |
    -52.5 ------+------- 52.5
                |
                |
               34.0
"""
from __future__ import annotations
from typing import Union
import math

from pyrusgeom.size_2d import Size2D
from pyrusgeom.segment_2d import Segment2D
from pyrusgeom.region_2d import Region2D
from pyrusgeom.ray_2d import Ray2D
from pyrusgeom.vector_2d import Vector2D
from pyrusgeom.line_2d import Line2D
from pyrusgeom.math_values import EPSILON


class Rect2D(Region2D):
    """ handling rectangles in SS2D
    Attributes:
        _top_left: a Vector2D for top left corner point
        _size: a Size2D for size value
        _is_valid: a boolean for validation
    """

    def __init__(self, *args) -> None:
        """This is the class init function for Rect2D.

        Defualt:
            bulid a Rect2D with args
            Even if argument point has incorrect values,
            the assigned values are normalized automatically.

        Raises:
            Exception: The input must be (4 float) or
            (vector2d and 2 float) or (vector2d and size2d) or (2 vector2d) or (1 rect2d)

        Args:
            four:
                four float:
                    float: left x
                    float: top y
                    float: length (x-range)
                    float: width (y-range)
            three:
                Vector and two float:
                    Vector2D: top left point
                    float: length X range
                    float: width Y range
            two:
                Vector, Size:
                    Vector2D: top left point
                    Size2D: size XY range
                two Vectors:
                    Vector2D: top left vertex
                    Vector2D: bottom right vertex
            one:
                Rect2D: just another rectangle
        """
        super().__init__()
        self._top_left: Vector2D = Vector2D()
        self._size: Size2D = Size2D()
        self._is_valid = True
        if len(args) == 4:
            self._top_left = Vector2D(args[0], args[1])
            self._size = Size2D(args[2], args[3])
            self._is_valid = True
        elif len(args) == 3:
            self._top_left = Vector2D(args[0])
            self._size = Size2D(args[1], args[2])
            self._is_valid = True
        elif len(args) == 2 and isinstance(args[1], Vector2D):
            self._top_left = Vector2D(args[0])
            bottom_right = Vector2D(args[1])
            top_left = Vector2D(args[0])
            self._size = Size2D(bottom_right.x() - top_left.x(),
                                bottom_right.y() - top_left.y())
            if bottom_right.x() - top_left.x() < 0.0:
                self._top_left.set_x(bottom_right.x())

            if bottom_right.y() - top_left.y() < 0.0:
                self._top_left.set_y(bottom_right.y())
                self._size = Size2D()
                self._is_valid = True
            else:
                self._top_left = Vector2D()
                self._size = Size2D()
                self._is_valid = True
        elif len(args) == 2:
            self._top_left = Vector2D(args[0])
            self._size = args[1]
            self._is_valid = True
        elif len(args) == 1 and isinstance(args[0], Rect2D):
            self._top_left = args[0].top_left()
            self._size = args[0].size()
            self._is_valid = args[0].is_valid()
        else:
            raise Exception("The input must be (4 float) or \
            (vector2d and 2 float) or (vector2d and size2d) or (2 vector2d) or (1 rect2d)")

    def assign(self, *args) -> Rect2D:
        """assign values directly

        Raises:
            Exception: The input must be (4 float) or
            (vector2d and 2 float) or (vector2d and size2d) or (2 vector2d) or (1 rect2d)

        Args:
            four:
                four float:
                    float: left x
                    float: top y
                    float: length (x-range)
                    float: width (y-range)
            three:
                Vector and two float:
                    Vector2D: top left point
                    float: length X range
                    float: width Y range
            two:
                Vector, Size:
                    Vector2D: top left point
                    Size2D: size XY range
                two Vectors:
                    Vector2D: top left vertex
                    Vector2D: bottom right vertex
            one:
                Rect2D: just another rectangle

        Returns:
            Rect2D: return self
        """

        if len(args) == 4:
            self._top_left.assign(args[0], args[1])
            self._size.assign(args[0], args[3])
        elif len(args) == 3:
            self._top_left = Vector2D(args[0])
            self._size.assign(args[1], args[2])
        elif len(args) == 2 and isinstance(args[1], Vector2D):
            self._top_left = Vector2D(args[0])
            bottom_right = Vector2D(args[1])
            top_left = Vector2D(args[0])
            self._size = Size2D(bottom_right.x() - top_left.x(),
                                bottom_right.y() - top_left.y())
            if bottom_right.x() - top_left.x() < 0.0:
                self._top_left.set_x(bottom_right.x())

            if bottom_right.y() - top_left.y() < 0.0:
                self._top_left.set_y(bottom_right.y())
                self._size = Size2D()
            else:
                self._top_left = Vector2D()
                self._size = Size2D()
        elif len(args) == 2:
            self._top_left = Vector2D(args[0])
            self._size = args[1]
        elif len(args) == 1 and isinstance(args[0], Rect2D):
            self._top_left = args[0].top_left()
            self._size = args[0].size()
        else:
            raise Exception("The input must be (4 float) or \
            (vector2d and 2 float) or (vector2d and size2d) or (2 vector2d)")
        return self

    def move_center(self, point: Vector2D) -> None:
        """move the center of rectangle to move whole rect.

        the center point is set to the given position.
        the size is unchanged.

        Args:
            point (Vector2D): center coordinates
        """
        self._top_left.assign(point.x() - self._size.length()
                              * 0.5, point.y() - self._size.width() * 0.5)

    def move_top_left(self, point: Vector2D) -> None:
        """move the top-left of rectangle to move whole rect.

        the top-left corner is set to the given position.
        the size is unchanged.

        Args:
            point (Vector2D): top-left corner
        """
        self._top_left = Vector2D(point)

    def move_bottom_right(self, point: Vector2D) -> None:
        """move the bottom-right of rectangle to move whole rect.

        the bottom-right corner is set to the given position.
        the size is unchanged.

        Args:
            point (Vector2D): bottom-right corner
        """
        self._top_left.assign(point.x() - self._size.length(),
                              point.y() - self._size.width())

    def move_left(self, left_x: float) -> None:
        """move the rectangle by moving left side

        the left line is set to the given position.
        the size is unchanged.

        Args:
            left_x (float): left side value
        """
        self._top_left.set_x(left_x)

    def move_min_x(self, left_x: float) -> None:
        """alias of moveLeft

        move the rectangle by moving left side.
        the left line is set to the given position.
        the size is unchanged.

        Args:
            left_x (float): left value
        """
        self.move_left(left_x)

    def move_right(self, right_x: float) -> None:
        """move the rectangle by moving left side(right given)

        the left line is set to the alterd position.
        the size is unchanged.

        Args:
            right_x (float): right value
        """
        self._top_left.set_x(right_x - self._size.length())

    def move_max_x(self, right_x: float) -> None:
        """alias of moveRight.

        move the rectangle by moving left side(right given).
        the left line is set to the alterd position.
        the size is unchanged.

        Args:
            right_x (float): right value
        """
        self.move_right(right_x)

    def move_top(self, top_y: float) -> None:
        """move the rectangle by moving top side

        the top line is set to the given value.
        the size is unchanged.

        Args:
            top_y (float): top value
        """
        self._top_left.set_y(top_y)

    def move_min_y(self, top_y: float) -> None:
        """alias of moveTop.

        move the rectangle by moving top side
        the top line is set to the given value.
        the size is unchanged.

        Args:
            top_y (float): top value
        """
        self.move_top(top_y)

    def move_bottom(self, bottom_y: float) -> None:
        """move the rectangle by moving top side(bottom given)

        the top line is set to the alterd value.
        the size is unchanged.

        Args:
            bottom_y (float): bottom value
        """
        self._top_left.set_y(bottom_y - self._size.width())

    def move_max_y(self, bottom_y: float) -> None:
        """alias of move_bottom.

        move the rectangle by moving top side(bottom given)
        the top line is set to the alterd value.
        the size is unchanged.

        Args:
            bottom_y (float): bottom value
        """
        self.move_bottom(bottom_y)

    def set_top_left(self, *args) -> Rect2D:
        """set the top-left corner of the rectangle.

        the size may be changed, the bottom-right corner will never be changed.

        Args:
            two:
                2 float:
                    float: x coordinate
                    float: y coordinate
            one:
                1 Vector:
                Vector2D: point coordinate
        Returns:
            Rect2D: this rect
        """
        point_x = 0.0
        point_y = 0.0
        if len(args) == 1:
            point_x = args[0].x()
            point_y = args[0].y()
        elif len(args) == 2:
            point_x = args[0]
            point_y = args[1]
        new_left = min(self.right(), point_x)
        new_right = max(self.right(), point_x)
        new_top = min(self.bottom(), point_y)
        new_bottom = max(self.bottom(), point_y)
        return self.assign(new_left, new_top, new_right - new_left, new_bottom - new_top)

    def set_bottom_right(self, *args) -> Rect2D:
        """set the bottom-right corner of the rectangle.

        the size may be changed, the top-left corner will never be changed.

        Args:
            two:
                2 float:
                    float: x coordinate
                    float: y coordinate
            one:
                1 Vector:
                Vector2D: point coordinate
        Returns:
            Rect2D: this rect
        """
        point_x = 0.0
        point_y = 0.0
        if len(args) == 1:
            point_x = args[0].x()
            point_y = args[0].y()
        elif len(args) == 2:
            point_x = args[0]
            point_y = args[1]
        new_left = min(self.left(), point_x)
        new_right = max(self.left(), point_x)
        new_top = min(self.top(), point_y)
        new_bottom = max(self.top(), point_y)
        return self.assign(new_left, new_top, new_right - new_left, new_bottom - new_top)

    def set_left(self, left_x: float) -> None:
        """set the left side of rectangle.

        the size may be changed, the right side will never be changed.

        Args:
            left_x (float): left value
        """
        new_left = min(self.right(), left_x)
        new_right = max(self.right(), left_x)
        self._top_left.set_x(new_left)
        self._size.set_length(new_right - new_left)

    def set_min_x(self, left_x: float) -> None:
        """alias of setLeft.

        set the left side of rectangle.
        the size may be changed, the right side will never be changed.

        Args:
            left_x (float): left value
        """
        self.set_left(left_x)

    def set_right(self, right_x: float) -> None:
        """set the right side of rectangle.

        the size may be changed, the left side will never be changed.

        Args:
            right_x (float): right value
        """
        new_left = min(self.left(), right_x)
        new_right = max(self.left(), right_x)

        self._top_left.set_x(new_left)
        self._size.set_length(new_right - new_left)

    def set_max_x(self, right_x: float) -> None:
        """ alias of setRight.

        set the right side of rectangle.
        the size may be changed, the left side will never be changed.

        Args:
            right_x (float): right value
        """
        self.set_right(right_x)

    def set_top(self, top_y: float) -> None:
        """set the top side of rectangle.

        the size may be changed, the bottom side will never be changed.

        Args:
            top_y (float): top value
        """
        new_top = min(self.bottom(), top_y)
        new_bottom = max(self.bottom(), top_y)

        self._top_left.set_y(new_top)
        self._size.set_width(new_bottom - new_top)

    def set_min_y(self, top_y: float) -> None:
        """alias of setTop.

        set the top side of rectangle.
        the size may be changed, the bottom side will never be changed.

        Args:
            top_y (float): top value
        """
        self.set_top(top_y)

    def set_bottom(self, bottom_y: float) -> None:
        """set the bottom side of rectangle.

        the size may be changed, the top side will never be changed.

        Args:
            bottom_y (float): bottom value
        """
        new_top = min(self.top(), bottom_y)
        new_bottom = max(self.top(), bottom_y)

        self._top_left.set_y(new_top)
        self._size.set_width(new_bottom - new_top)

    def set_max_y(self, bottom_y: float) -> None:
        """alias of setBottom.

        set the bottom side of rectangle.
        the size may be changed, the top side will never be changed.

        Args:
            bottom_y (float): bottom value
        """
        self.set_bottom(bottom_y)

    def set_length(self, length: float) -> None:
        """set a x-range - size x

        Args:
            length (float): length range
        """
        self._size.set_length(length)

    def set_width(self, width: float) -> None:
        """set a y-range - size y

        Args:
            length (float): width range
        """
        self._size.set_width(width)

    def set_size(self, *args) -> None:
        """set size of rect

        Raises:
            Exception: input must be two float or one size

        Args:
            two:
                2 float:
                    float: length range
                    float: width range
            one:
                1 size:
                    Size2D: size range
        """
        if len(args) == 2:
            self._size.assign(args[0], args[1])
        elif len(args) == 1 and isinstance(args[0],Size2D):
            self._size = Size2D(args[0].length(),args[0].width())
        else:
            raise Exception("input must be two float or one size")

    def is_valid(self) -> bool:
        """check if self rectangle is valid or not.

        Returns:
            bool: True if the area of self rectangle is not 0. else False
        """
        self._is_valid = self._size.is_valid()
        return self._is_valid

    def area(self) -> float:
        """get the area value of self rectangle.

        Returns:
            float: value of the area
        """
        return self._size.length() * self._size.width()

    def contains(self, point: Vector2D) -> bool:
        """check if point is within self region.

        Args:
            point (Vector2D): considered point

        Returns:
            bool: True if it contains it. else False.
        """
        return self.left() <= point.x() <= self.right() and self.top() <= point.y() <= self.bottom()

    def contains_almost(self, point, error_thr) -> bool:
        """check if point is within self region with error threshold.

        Args:
            point ([type]): considered point
            error_thr ([type]): error threshold

        Returns:
            bool: True if it almost contains it. else False.
        """
        return self.left() - error_thr <= point.x <= self.right() + error_thr and \
            self.top() - error_thr <= point.y <= self.bottom() + error_thr

    def left(self) -> float:
        """get the left x coordinate of this rectangle.

        Returns:
            float: x coordinate value
        """
        return self._top_left.x()

    def right(self) -> float:
        """get the right x coordinate of this rectangle.

        Returns:
            float: x coordinate value
        """
        return self.left() + self._size.length()

    def top(self) -> float:
        """get the top y coordinate of this rectangle.

        Returns:
            float: y coordinate value
        """
        return self._top_left.y()

    def bottom(self) -> float:
        """get the bottom y coordinate of this rectangle.

        Returns:
            float: y coordinate value
        """
        return self.top() + self._size.width()

    def min_x(self) -> float:
        """get minimum value of x coordinate of this rectangle

        Returns:
            float: x coordinate value (equivalent to left())
        """
        return self.left()

    def max_x(self) -> float:
        """get maximum value of x coordinate of this rectangle

        Returns:
            float: x coordinate value (equivalent to right())
        """
        return self.right()

    def min_y(self) -> float:
        """get minimum value of y coordinate of this rectangle

        Returns:
            float: y coordinate value (equivalent to top())
        """
        return self.top()

    def max_y(self) -> float:
        """get maximum value of y coordinate of this rectangle

        Returns:
            float: y coordinate value (equivalent to bottom())
        """
        return self.bottom()

    def size(self) -> Size2D:
        """get the XY copy range of self rectangle

        Returns:
            Size2D: new size object
        """
        return Size2D(self._size.length(),self._size.width())

    def size_(self) -> Size2D:
        """get the OG XY range of self rectangle

        Returns:
            Size2D: original size object
        """
        return self._size

    def center(self) -> Vector2D:
        """get center point of this rectangle

        Returns:
            Vector2D: coordinate value by vector object
        """
        return Vector2D((self.left() + self.right()) * 0.5,
                        (self.top() + self.bottom()) * 0.5)

    def top_left(self) -> Vector2D:
        """get the top-left corner copy vector point

        Returns:
            Vector2D: coordinate value by vector object copy
        """
        return Vector2D(self._top_left)

    def top_left_(self) -> Vector2D:
        """get the og top-left corner vector point

        Returns:
            Vector2D: coordinate value by og vector object
        """
        return self._top_left

    def top_right(self) -> Vector2D:
        """get the top-right corner point

        Returns:
            Vector2D: coordinate value by vector object
        """
        return Vector2D(self.right(), self.top())

    def bottom_left(self) -> Vector2D:
        """get the bottom-left corner point

        Returns:
            Vector2D: coordinate value by vector object
        """
        return Vector2D(self.left(), self.bottom())

    def bottom_right(self) -> Vector2D:
        """get the bottom-right corner point

        Returns:
            Vector2D: coordinate value by vector object
        """
        return Vector2D(self.right(), self.bottom())

    def left_edge(self) -> Line2D:
        """get the left edge line

        Returns:
            Line2D: line object
        """
        return Line2D(self.top_left(), self.bottom_left())

    def right_edge(self) -> Line2D:
        """get the right edge line

        Returns:
            Line2D: line object
        """
        return Line2D(self.top_right(), self.bottom_right())

    def top_edge(self) -> Line2D:
        """get the top edge line

        Returns:
            Line2D: line object
        """
        return Line2D(self.top_left(), self.top_right())

    def bottom_edge(self) -> Line2D:
        """get the bottom edge line

        Returns:
            Line2D: line object
        """
        return Line2D(self.bottom_left(), self.bottom_right())

    def intersection(self, other:Union[Line2D,Ray2D,Segment2D]) -> list[Vector2D]:
        """calculate intersection point with line.
        Raises:
            Exception: Input must be Line/Ray/Segment
        Args:
            other (Union[Line2D,Ray2D,Segment2D]):
                Line2D:  considered line.
                Ray2D: considered ray line.
                Segment2D: considered line segment.

        TODO: check this full

        Returns:
            list[Vector2D]: intersection Points
        """
        if isinstance(other, Line2D):
            n_sol = 0
            t_sol = [Vector2D(0, 0), Vector2D(0, 0)]
            left_x = self.left()
            right_x = self.right()
            top_y = self.top()
            bottom_y = self.bottom()
            t_sol[n_sol] = self.left_edge().intersection(other)
            if n_sol < 2 and t_sol[n_sol].is_valid() and top_y <= t_sol[n_sol].y() <= bottom_y:
                n_sol += 1
            t_sol[n_sol] = self.right_edge().intersection(other)
            if n_sol < 2 and t_sol[n_sol].is_valid() and top_y <= t_sol[n_sol].y() <= bottom_y:
                n_sol += 1
            t_sol[n_sol] = self.top_edge().intersection(other)
            if n_sol < 2 and (t_sol[n_sol]).is_valid() and left_x <= t_sol[n_sol].x() <= right_x:
                n_sol += 1
            t_sol[n_sol] = self.top_edge().intersection(other)
            if n_sol < 2 and (t_sol[n_sol]).is_valid() and left_x <= t_sol[n_sol].x() <= right_x:
                n_sol += 1
            if n_sol == 2 and math.fabs(t_sol[0].x() - t_sol[1].x()) < EPSILON and math.fabs(
                    t_sol[0].y() - t_sol[1].y()) < EPSILON:
                n_sol = 1
            if n_sol == 0:
                sol_list = []
            elif n_sol == 1:
                sol_list = [t_sol[0]]
            else:
                sol_list = [t_sol[0], t_sol[1]]
            return sol_list
        if isinstance(other, Ray2D):
            n_sol = self.intersection(other.line())

            if len(n_sol) > 1 and not other.in_right_dir(n_sol[2], 1.0):
                del n_sol[1]

            if len(n_sol) > 0 and not other.in_right_dir(n_sol[1], 1.0):
                n_sol[0] = n_sol[1]
                del n_sol[1]
            return n_sol
        if isinstance(other, Segment2D):
            n_sol = self.intersection(other.line())
            if len(n_sol) > 1 and not other.contains(n_sol[2]):
                del n_sol[1]

            if len(n_sol) > 0 and not other.contains(n_sol[1]):
                n_sol[0] = n_sol[1]
                del n_sol[1]
            return n_sol
        raise Exception("Input must be Line/Ray/Segment")

    def intersected(self, other:Rect2D) -> Rect2D:
        """get the intersected rectangle of self rectangle and the other rectangle.

        If no intersection between rectangles,empty rectangle is returned.

        Args:
            other (Rect2D): other rectangle

        Returns:
            [Rect2D]: rectangle instance
        """
        if not self.is_valid or not other.is_valid():
            self._top_left.assign(0.0, 0.0)
            self._size.assign(0.0, 0.0)

        left = max(self.left(), other.left())
        top = max(self.top(), other.top())
        width = min(self.right(), other.right()) - left
        length = min(self.bottom(), other.bottom()) - top

        if width <= 0.0 or length <= 0.0:
            self._top_left.assign(0.0, 0.0)
            self._size.assign(0.0, 0.0)

        self._top_left.assign(left, top)
        self._size.assign(width, length)

        return self

    def united(self, other:Rect2D) -> Rect2D:
        """get the united rectangle of self rectangle and the other rectangle.

        Args:
            other (Rect2D): other rectangle

        Returns:
            Rect2D: rectangle instance.
        """
        if not self.is_valid or not other.is_valid():
            self._top_left.assign(0.0, 0.0)
        self._size.assign(0.0, 0.0)

        left = max(self.left(), other.left())
        top = max(self.top(), other.top())
        width = min(self.right(), other.right()) - left
        length = min(self.bottom(), other.bottom()) - top

        if width <= 0.0 or length <= 0.0:
            self._top_left.assign(0.0, 0.0)
            self._size.assign(0.0, 0.0)

        self._top_left.assign(left, top)
        self._size.assign(width, length)

        return self

    @staticmethod
    def from_center(*args)->Rect2D:
        """create rectangle with center point and size.

        Raises:
            Exception: Input must be 4 float or one vector and two float

        Args:
            four:
                4 float:
                    float: x value of center point of rectangle.
                    float: y value of center point of rectangle.
                    float: length(x-range) of rectangle.
                    float: width(y-range) of rectangle.
            three:
                1 vector and 2 float:
                    Vector2D: center point of rectangle.
                    float: length(x-range) of rectangle.
                    float: width(y-range) of rectangle.

        Returns:
            Rect2D: created rectangle
        """
        if len(args) == 4:
            return Rect2D(args[0] - args[2] * 0.5, args[1] - args[3] * 0.5, args[2], args[3])
        if len(args) == 3:
            return Rect2D(args[0] - args[1] * 0.5, args[0] - args[2] * 0.5, args[1], args[2])
        raise Exception("Input must be 4 float or one vector and two float")

    def __repr__(self) -> str:
        """represent Rect2D as a string

        Returns:
            str: Rect2D's top_left corner and length and width as string
        """
        return f"[topleft:{self._top_left},len:{self._size.length()},wid:{self._size.width()}]"

    def to_str(self, ostr) -> str:
        """add rectangle to the ostr

        Args:
            ostr (str): str to add to it
        """
        ostr += f'(rect {round(self.left(), 3)} {round(self.top(), 3)} \
            f{round(self.right(), 3)} {round(self.bottom(), 3)})'



""" ray_2d.py file
    Ray2D: class name
    Class attributes:_origin,_drirection,_is_valid
"""
from __future__ import annotations
from typing import Union
import math

from pyrusgeom.line_2d import Line2D
from pyrusgeom.vector_2d import Vector2D
from pyrusgeom.angle_deg import AngleDeg
from pyrusgeom.math_values import EPSILON


class Ray2D:
    """ handling rays in SS2D
    Attributes:
        _origin: a Vector2D for the orgin point
        _direction: an AngleDeg for the driection the ray goes
        _is_valid = a boolean for validation
    """

    def __init__(self, origin: Vector2D, direction: Union[Vector2D, AngleDeg, float, int]):
        """This is the class init function for Ray2D.

        Defualt:
            create a Ray started at origin point which has a angle at _direction

        Args:
            origin (Vector2D): origin point
            direction (Union[Vector2D,AngleDeg, float, int]): direction angle

        Raises:
            Exception: The input must be (two Vector2D) or (Vector2D and AngleDeg)
        """
        self._origin: Vector2D = Vector2D(origin)
        self._direction: AngleDeg = AngleDeg()
        self._is_valid = False
        if isinstance(direction, Vector2D):
            self._direction = (direction - origin).th()
            self._is_valid = True
        elif isinstance(direction, (AngleDeg, float, int)):
            self._direction = AngleDeg(direction)
            self._is_valid = True
        elif not self._is_valid:
            raise Exception(
                'The input should be (two Vector2D) or (A Vector2D and an AngleDeg)')

    def origin(self) -> Vector2D:
        """get origin point copy

        Returns:
            Vector2D: origin point value
        """
        return Vector2D(self._origin)

    def dir(self) -> AngleDeg:
        """get the angle copy of this ray line

        Returns:
            AngleDeg: direaction value
        """
        return AngleDeg(self._direction)

    def origin_(self) -> Vector2D:
        """get origin point reference

        Returns:
            Vector2D: og origin point
        """
        return self._origin

    def dir_(self) -> AngleDeg:
        """get the reference direaction of this ray

        Returns:
            AngleDeg: og direaction
        """
        return self._direction

    def copy(self) -> Ray2D:
        """create a copy of this ray

        Returns:
            Ray2D: copy of og ray
        """
        return Ray2D(self.origin(), self.dir())

    def line(self) -> Line2D:
        """get line generated from this ray

        Returns:
            Line2D: new line object
        """
        return Line2D(self._origin, self._direction)

    def in_right_dir(self, point: Vector2D, thr: float = 10.0) -> bool:
        """check whether point is on the direction of this Ray

        Args:
            point (Vector2D): considered point
            thr (float, optional): threshold angle buffer. Defaults to 10.0.

        Returns:
            bool: true if it is on the direction. else false.
        """
        return ((point - self._origin).th() - self._direction).abs() < thr

    def intersection(self, other: Union[Line2D, Ray2D]) -> Vector2D:
        """get the intersection point with line or ray

        Args:
            other (Union[Line2D,Ray2D]): considered line

        Returns:
            Vector2D: intersection point. if it does not exist,
                    the invalidated value vector is returned.
        Raises:
            Exception: The input must be Line2D or Ray2D
        """
        if isinstance(other, Ray2D):
            tmp_sol = self.line().intersection(other.line())
            if not tmp_sol.is_valid():
                return Vector2D.invalid()

            if not self.in_right_dir(tmp_sol) or not other.in_right_dir(tmp_sol):
                return Vector2D.invalid()

            return tmp_sol
        if isinstance(other, Line2D):
            tmp_sol = self.line().intersection(other)
            if not tmp_sol.is_valid():
                return Vector2D.invalid()

            if not self.in_right_dir(tmp_sol):
                return Vector2D.invalid()

            return tmp_sol

        raise Exception("The input should be Line2D or Ray2D")

    def __eq__(self, other: Ray2D) -> bool:
        """operator == for Ray2D

        Args:
            other (Ray2D): right hand side argument
        Returns:
            bool: true if equal or difference is less than EPSILON. else false
        """
        return (math.fabs(self._direction - other.dir_()) < EPSILON
                and self._origin == other.origin_())

    def __repr__(self) -> str:
        """represent Ray2D as a string

        Returns:
            str: Ray2D's origin and direction as string
        """
        return str(self._origin) + " dir : " + str(self._direction)



""" convex_hull.py file
    ConvexHull: class name
"""
from __future__ import annotations
from enum import Enum, unique, auto
import functools
import math

from pyrusgeom.triangle_2d import Triangle2D
from pyrusgeom.polygon_2d import Polygon2D
from pyrusgeom.vector_2d import Vector2D
from pyrusgeom.segment_2d import Segment2D
from pyrusgeom.math_values import EPSILON


@unique
class MethodType(Enum):
    """ MethodType
        Algorithm types:
            DIRECT_METHOD: don't use it need fixing
            WRAPPING_METHOD
            GRAHAN_SCAN
    """
    DIRECT_METHOD = auto()
    WRAPPING_METHOD = auto()
    GRAHAN_SCAN = auto()


def list2set(list_: list) -> list:
    """function to convert a list to a set with unique values

    Args:
        list_ (list): list with duplicate items

    Returns:
        list: a set with unique values
    """
    sub = []
    for item in list_:
        # check if exists in sub or not
        if item not in sub:
            sub.append(item)
    return sub

# Base for angle_sort_predicate
# base = Vector2D()


class AngleSortPredicate:
    """angle_sort_predicate cmp key for sort
    """

    def __init__(self, obj, base_in, *args):
        self.obj = obj
        self.base_in = base_in
        self.args = args

    def __lt__(self, other):
        return angle_sort_predicate(self.obj, other.obj, base=self.base_in) < 0

    def __gt__(self, other):
        return angle_sort_predicate(self.obj, other.obj, base=self.base_in) > 0

    def __eq__(self, other):
        return angle_sort_predicate(self.obj, other.obj, base=self.base_in) == 0

    def __le__(self, other):
        return angle_sort_predicate(self.obj, other.obj, base=self.base_in) <= 0

    def __ge__(self, other):
        return angle_sort_predicate(self.obj, other.obj, base=self.base_in) >= 0

    def __ne__(self, other):
        return angle_sort_predicate(self.obj, other.obj, base=self.base_in) != 0


def angle_sort_predicate(lhs: Vector2D, rhs: Vector2D, base: Vector2D = Vector2D()) -> int:
    """compare and check the base[(0,0) by default] and lhs and rhs to find are they
        in a counter clockwise order or not
        base - lhs - rhs

    Args:
        lhs (Vector2D): first point
        rhs (Vector2D): second point
        base (Vector2D): Base for angle_sort_predicate

    Returns:
        bool: if counter clockwise return 1. else -1
    """
    area = Triangle2D.double_signed_area_st(base, lhs, rhs)

    if area < 0.0:
        return -1
    if area < EPSILON:
        if base.get_y() < lhs.get_y():
            if base.dist2(lhs) > base.dist2(rhs):
                return -1
        else:
            if base.dist2(lhs) < base.dist2(rhs):
                return -1
    return 1


def is_clockwise(point0: Vector2D, point1: Vector2D, point2: Vector2D) -> bool:
    """compare and check the point0 and point1 and point2 to find are they
        in a clockwise order or not

    Args:
        point0 (Vector2D): first point
        point1 (Vector2D): second point
        point2 (Vector2D): third point

    Returns:
        bool: if clockwise return true. else false
    """
    area = Triangle2D.double_signed_area_st(point0, point1, point2)

    return area < 0.0 or (area < EPSILON and point0.dist2(point1) > point0.dist2(point2))


class ConvexHull:
    """ handling the Convex Hull in SS2D
    Attributes:
        _input_points: input points
        _vertices: vertices of convex hull [counter clockwise order]
        _edges: edges of convex hull
    """

    def __init__(self, point_list: list):
        """This is the class init function and creates the convex hull.

        Defualt:
            create empty convex hull
        OR
            create convex hull with given points
        Args:
            point_list: array of input points

        """
        self._input_points = point_list
        self._vertices = []
        self._edges = []

    def clear(self) -> None:
        """clear all.
        """
        self.clear_results()
        self._input_points.clear()

    def clear_results(self) -> None:
        """clear result variables.
        """
        self._vertices.clear()
        self._edges.clear()

    def add_point(self, point: Vector2D) -> None:
        """add a point to the set of input point

        Args:
            point(Vector2D) : point to add
        """
        self._input_points.append(point)

    def add_points(self, points: list) -> None:
        """add a point list to the set of input point

        Args:
            points(list) : points to add
        """
        for point in points:
            self._input_points.append(point)

    def compute(self, m_type: MethodType = MethodType.WRAPPING_METHOD) -> None:
        """generate convex hull by specified method
            DIRECT_METHOD: don't use it need fixing
            WRAPPING_METHOD
            GRAHAN_SCAN
        Args:
            m_type(MethodType):the specified method id

        """
        if m_type == MethodType.DIRECT_METHOD:
            self.compute_direct_method()
        elif m_type == MethodType.GRAHAN_SCAN:
            self.compute_graham_scan()
        elif m_type == MethodType.WRAPPING_METHOD:
            self.compute_wrapping_method()

    def input_points(self) -> list:
        """get a copy list from the input point container

        Returns:
            list: a list with input point conrainer values
        """
        input_list_cp = self._input_points.copy()
        return input_list_cp

    def input_points_(self) -> list:
        """ get the reference and og list to the input point container

        Returns:
            list: a reference to the input point container
        """
        return self._input_points

    def vertices(self) -> list[Vector2D]:
        """get a copy list from the vertex container

        Returns:
            list: a list with the vertex container values
        """
        vertices_cp = self._vertices.copy()
        return vertices_cp

    def vertices_(self) -> list:
        """ get the reference and og list to the vertex container

        Returns:
            list: a reference to the vertex container
        """
        return self._vertices

    def edges(self) -> list:
        """get a copy list from the edge container

        Returns:
            list: a list with the edge container values
        """
        edges_cp = self._edges.copy()
        return edges_cp

    def edges_(self) -> list:
        """ get the reference and og list to the edge container

        Returns:
            list: a reference to the edge container
        """
        return self._edges

    def compute_direct_method(self):
        """direct method
            don't use this one until the fix
        """
        self.clear_results()

        point_size = len(self._input_points)

        if point_size < 3:
            return

        for i in range(point_size - 1):
            point_first:Vector2D = self._input_points[i]
            for j in range(i + 1, point_size):
                point_second:Vector2D = self._input_points[j]
                rel = point_second - point_first

                valid = True
                last_value = 0.0

                for k in range(point_size):
                    if k in (i, j):
                        continue

                    point_third = self._input_points[k]
                    outer_prod = rel.outerProduct(point_third - point_first)

                    if math.fabs(outer_prod) < EPSILON:
                        # point is on the line
                        if (point_third - point_first).r2() < rel.r2():
                            # point is on the segment
                            valid = False
                            break
                    # to check if other_prod and last_value don't have a same
                    # sign means one of the points exists in the opposite side
                    if outer_prod ^ last_value < 0.0:
                        valid = False
                        break

                    last_value = outer_prod

                if valid:
                    self._vertices.append(point_first)
                    self._vertices.append(point_second)

                    if last_value < 0.0:
                        self._edges.append(
                            Segment2D(point_first, point_second))

                    else:
                        self._edges.append(
                            Segment2D(point_second, point_first))

        # sort vertices by counter clockwise order

        if len(self._vertices) > 0:
            self._vertices.sort(key=functools.cmp_to_key(angle_sort_predicate))
            # TODO: Need to pass the base to the cmp-key fnction
            # base is the line bellow the last member of _vertices
            # until then avoid this method. we are open to
            # any suggetion for solving this cmp-key problem
            # base = self._vertices[len(self._vertices) - 1]
            # functools.cmp_to_key(angle_sort_predicate)
            # AngleSortPredicate(base_in=base)
            # (base:self._vertices[len(self._vertices) - 1])

            self._vertices = list2set(self._vertices)

    def compute_wrapping_method(self):
        """wrapping method
        """
        self.clear_results()

        point_size = len(self._input_points)
        min_index = self.get_min_point_index()

        if point_size < 3 or min_index == -1:
            return

        vertices_index = []  # temporal set for checking already used vertices.

        self._vertices.append(self._input_points[min_index])

        current_index = min_index
        current_point:Vector2D = self._input_points[min_index]
        for _ in range(point_size + 1):
            candidate = 0
            for i in range(point_size):
                if i == current_index or i in vertices_index:
                    continue

                candidate = i
                break
            for i in range(candidate + 1, point_size):
                if i == current_index or i in vertices_index:
                    continue

                point_first:Vector2D = self._input_points[candidate]
                point_second:Vector2D = self._input_points[i]

                area = Triangle2D.double_signed_area_st(
                    current_point, point_first, point_second)

                if area < 0.0:
                    candidate = i

                elif (area < EPSILON and
                        current_point.dist2(point_first) > current_point.dist2(point_second)):
                    candidate = i
            current_index = candidate
            current_point = self._input_points[current_index]
            vertices_index.append(current_index)
            self._vertices.append(current_point)

            if current_index == min_index:
                break
        for i in range(len(self._vertices)):
            if self._vertices[i] != self._vertices[-1]:
                self._edges.append(
                    Segment2D(self._vertices[i], self._vertices[i + 1]))

        self._vertices.pop()

    def compute_graham_scan(self):
        """Graham scan method
        """
        self.clear_results()

        point_size = len(self._input_points)
        min_index = self.get_min_point_index()

        if point_size < 3 or min_index == -1:
            return

        self.sort_points_angle_from(min_index)

        self._vertices = self._input_points

        top = 1
        for i in range(2, point_size):
            while not is_clockwise(self._vertices[top - 1],
                self._vertices[top], self._input_points[i]):
                top -= 1

            top += 1
            tmp = self._vertices[top]
            self._vertices[top] = self._vertices[i]
            self._vertices[i] = tmp

        top += 1

        del self._vertices[top:len(self._vertices)]

        for i in range(len(self._vertices)):
            if self._vertices[i] != self._vertices[-1]:
                self._edges.append(
                    Segment2D(self._vertices[i], self._vertices[i + 1]))
        self._edges.append(Segment2D(self._vertices[0], self._vertices[-1]))

    def get_min_point_index(self) -> int:
        """get the index of minimum coordinate point

        Returns:
            int: if there is no point return -1 else index of minimum coord
        """
        point_size = len(self._input_points)

        if point_size == 0:  # or point_size < 3:
            return -1

        min_index = 0

        min_point: Vector2D = self._input_points[0]
        for i in range(point_size):
            point: Vector2D = self._input_points[i]

            if min_point.x() > point.x() or (min_point.x() == point.x() and
                                             min_point.y() > point.y()):
                min_point = point
                min_index = i

        return min_index

    def sort_points_angle_from(self, index):
        """sort points by anlge from base(0,0)
        """
        if len(self._input_points) <= index:
            return

        tmp = self._input_points[0]
        self._input_points[0] = self._input_points[index]
        self._input_points[index] = tmp

        # base is (0,0)
        self._input_points.sort(key=functools.cmp_to_key(angle_sort_predicate))

    def to_polygon(self) -> Polygon2D:
        """create and get the convex hull polygon

        Returs:
            Polygon2D: a ploygon from the convex hull _vertices
        """
        return Polygon2D(self._vertices)

    def __repr__(self) -> str:
        """represent the convex hull as a string

        Returns:
            str: contains _vertices and _edges
        """
        return f"({self._vertices} , {self._edges})"


# def test():
#     point_list = [Vector2D(), Vector2D(1, 1), Vector2D(0, 1)]
#     c = ConvexHull(point_list)
#     print(c)



""" angle_deg.py file
    AngleDeg : class name
    Class attributes : degree
"""
from __future__ import annotations
from typing import Union
import math
from pyrusgeom.math_values import RAD2DEG,EPSILON,DEG2RAD

class AngleDeg:
    """ handling degrees in SS2D
    Attributes:
        degree: a float for angles
    """
    def __init__(self, *args, **kwargs) -> None:
        """This is the class init function and normalizes the input degree.

        Defualt:
            bulid a AngleDeg with value of 0 degree
        Args:
            *args:
                none: for default angle
                one:
                    AngleDeg: create a new AngleDeg with normalized input AngleDeg
                    Degree: create an AngleDeg with normalized degree
        KWArgs:
            af (float): create an AngleDeg with normalized degree
        Raises:
            Exception: Input must be an AngleDeg or a Degree
        """
        if len(args) == 0 and len(kwargs) == 0:
            self._degree = 0
        elif len(args) == 1:
            if isinstance(args[0], AngleDeg):
                self._degree = args[0].degree()
            else:
                self._degree = args[0]
        elif 'af' in kwargs:
            self._degree = kwargs['af']
        else:
            raise Exception('The input should be an AngleDeg or a Degree')
        self.normal()

    def degree(self) -> float:
        """get the degree in a new AngleDeg copy
        # TODO: do we need this?
        Returns:
            float: a new AngleDeg degree with the old AngleDeg's degree
        """
        degree_cp = AngleDeg(self._degree)
        return degree_cp.degree_()

    def degree_(self) -> float:
        """get the orginal degree

        Returns:
            float: AngleDeg's degree
        """
        return self._degree

    def set_degree(self, degree: float) -> None:
        """set a new degree for AngleDeg and normalizes it.

        Args:
            degree (float): new degree to set
        """
        self._degree = degree
        self.normal()

    def set_angle(self, angle: AngleDeg) -> None:
        """set a new degree for AngleDeg from another AngleDeg

        Args:
            angle (AngleDeg): an AngleDeg to copy the degree
        """
        self._degree = angle.degree()

    def normal(self) -> None:
        """normalizing the degree

        if the angle is below -360 degree or above 360 degree,
        it will calculating module of given degree by 360,
        and cap the angle between -180 degree and 180 degree.

        """
        if self._degree < -360.0 or 360.0 < self._degree:
            self._degree = math.fmod(self._degree, 360.0)

        if self._degree < -180.0:
            self._degree += 360.0

        if self._degree > 180.0:
            self._degree -= 360.0

    def is_within(self, left: Union[AngleDeg, float, int],
        right: Union[AngleDeg, float, int]) -> bool:
        """check if this AngleDeg is within [left, right] (turn clockwise)

        Args:
            left (Union[AngleDeg, float, int]): left angle
            right (Union[AngleDeg, float, int]): right angle

        Returns:
            bool: true if this is within [left, right]. else false
        """
        if not isinstance(left, AngleDeg):
            left = AngleDeg(left)
        if not isinstance(right, AngleDeg):
            right = AngleDeg(right)
        if left.is_left_equal_of(right):
            if left.is_left_equal_of(self) and self.is_left_equal_of(right):
                return True
        else:
            if self.is_left_equal_of(right) or left.is_left_equal_of(self):
                return True
        return False

    def abs(self) -> float:
        """get absolute value of this AngleDeg

        Returns:
            float: absolute value of degree()
        """
        return math.fabs(self.degree())

    def radian(self) -> float:
        """get RADIAN value.

        Returns:
            float:  radian value of degree()
        """
        return self.degree() * DEG2RAD

    def copy(self) -> AngleDeg:
        """ copy the AngleDeg

        Returns:
            AngleDeg: a new AngleDeg with same values
        """
        return AngleDeg(self._degree)

    def reverse(self) -> None:
        """reverse this AngleDeg - 180 degrees opposite
        """
        if self._degree >= 0:
            self._degree = -(180 - self._degree)
        else:
            self._degree = 180 + self._degree

    def reverse_angle(self) -> AngleDeg:
        """make a reversed AngleDeg - 180 degrees opposite

        Returns:
            AngleDeg: a new AngleDeg with reverse values
        """
        angle = AngleDeg(self.degree())
        angle.reverse()
        return angle

    def get_normalized(self) -> float:
        """get the Normalized degree betwhen 0 and 1

        Returns:
            float: Normalized degree betwhen 0 and 1
        """
        return (self.degree() + 180) / 360

    def is_left_of(self, angle: Union[AngleDeg, float, int]) -> bool:
        """check if this AngleDeg is left of [angle]

        Args:
            angle (Union[AngleDeg, float, int]): angle to check

        Returns:
            bool: true if this AngleDeg is left of [angle]. else false
        """
        if not isinstance(angle, AngleDeg):
            angle = AngleDeg(angle)
        diff = angle.degree() - self.degree()
        return (0.0 < diff < 180.0) or diff < -180.0

    def is_right_of(self, angle: Union[AngleDeg, float, int]) -> bool:
        """check if this AngleDeg is right of [angle]

        Args:
            angle (Union[AngleDeg, float, int]): angle to check

        Returns:
            bool: true if this AngleDeg is right of [angle]. else false
        """
        if not isinstance(angle, AngleDeg):
            angle = AngleDeg(angle)
        diff = self.degree() - angle.degree()
        return (0.0 < diff < 180.0) or diff < -180.0

    def is_left_equal_of(self, angle: Union[AngleDeg, float, int]) -> bool:
        """check if this AngleDeg is left or equal of [angle]

        Args:
            angle (Union[AngleDeg, float, int]): angle to check

        Returns:
            bool: true if this AngleDeg is left of [angle]. else false
        """
        if not isinstance(angle, AngleDeg):
            angle = AngleDeg(angle)
        diff = angle.degree() - self._degree
        return 0.0 <= diff < 180.0 or diff < -180.0

    def is_right_equal_of(self, angle: Union[AngleDeg, float, int]) -> bool:
        """check if this AngleDeg is right or equal of [angle]

        Args:
            angle (Union[AngleDeg, float, int]): angle to check

        Returns:
            bool: true if this AngleDeg is right of [angle]. else false
        """
        if not isinstance(angle, AngleDeg):
            angle = AngleDeg(angle)
        diff = self.degree() - angle.degree()
        return (0.0 <= diff < 180.0) or diff < -180.0

    def __iadd__(self, other: Union[AngleDeg, float, int]) -> AngleDeg:
        """operator +=

        Args:
            other (Union[AngleDeg, float, int]): angle added value

        Returns:
            AngleDeg: this AngleDeg
        """
        if isinstance(other, AngleDeg):
            other = other.degree()
        self._degree += other
        self.normal()
        return self

    def __isub__(self, other: Union[AngleDeg, float, int]) -> AngleDeg:
        """operator -=

        Args:
            other (Union[AngleDeg, float, int]): angle subtract argument

        Returns:
            AngleDeg: this AngleDeg
        """
        if isinstance(other, AngleDeg):
            other = other.degree()
        self._degree -= other
        self.normal()
        return self

    def __imul__(self, other: Union[float, int]) -> AngleDeg:
        """operator *=

        Args:
            other (Union[float, int]): scalar multiply argument

        Returns:
            AngleDeg: this AngleDeg
        """
        self._degree *= other
        self.normal()
        return self

    def __idiv__(self, other: Union[float, int]) -> AngleDeg:
        """operator /=

        Args:
            other (Union[float, int]): scalar division argument

        Returns:
            AngleDeg: this AngleDeg
        """
        self._degree /= other
        self.normal()
        return self

    def __add__(self, other: Union[AngleDeg, float, int]) -> AngleDeg:
        """operator add for AngleDeg

        Args:
            other (Union[AngleDeg, float, int]): right hand side argument

        Returns:
            AngleDeg: a new Angledeg with value of
            sum of left hand side argument and right hand side argument
        """
        if isinstance(other,AngleDeg):
            new_angle_deg = AngleDeg(self._degree + other.degree())
        else:
            new_angle_deg = AngleDeg(self._degree + other)
        return new_angle_deg

    def __sub__(self, other: Union[AngleDeg, float, int]) -> AngleDeg:
        """operator sub for AngleDeg

        Args:
            other (Union[AngleDeg, float, int]): right hand side argument

        Returns:
            AngleDeg: a new Angledeg with value of
            subtraction of right hand side argument from left hand side argument
        """
        if isinstance(other,AngleDeg):
            return AngleDeg(self._degree - other._degree)
        return AngleDeg(self._degree - other)

    def __mul__(self, other: Union[float, int]) -> AngleDeg:
        """operator mul for AngleDeg

        Args:
            other (Union[float, int]): scalar multiply argument

        Returns:
            AngleDeg: a new AngleDeg with the multiply of this AngleDeg and
            scalar multiply argument
        """
        new_angle_deg = AngleDeg(self._degree * other)
        return new_angle_deg

    def __floordiv__(self, other: Union[float, int]) -> AngleDeg:
        """operator floordiv for AngleDeg

        Args:
            other (Union[float, int]): scalar division argument

        Returns:
            AngleDeg: a new AngleDeg with the floor division of this AngleDeg
            on scalar division argument
        """
        new_angle_deg = AngleDeg(self._degree / other)
        return new_angle_deg

    def __repr__(self) -> str:
        """represent AngleDeg as a string

        Returns:
            str: AngleDeg's Degree as string
        """
        return str(self.degree())

    def __float__(self) -> float:
        """represent AngleDeg as a float

        Returns:
            float: AngleDeg's Degree as float
        """
        return float(self.degree())

    def __neg__(self) -> AngleDeg:
        """ negation operator

        Returns:
            AngleDeg: new AngleDeg with negation
        """
        new_angle_deg = AngleDeg(-self._degree)
        return new_angle_deg

    def __eq__(self, other: Union[AngleDeg, float, int]) -> bool:
        """operator == for AngleDeg

        Args:
            other (Union[AngleDeg,float, int]): right hand side argument
        Returns:
            bool: true if equal or difference is less than EPSILON. else false
        """
        if isinstance(other, AngleDeg):
            other = other.degree()
        return math.fabs(self._degree - other) < EPSILON

    def cos(self) -> float:
        """calculate cosine

        Returns:
            float: cosine value
        """
        return math.cos(self._degree * DEG2RAD)

    def sin(self) -> float:
        """calculate sine

        Returns:
            float: sine value
        """
        return math.sin(self._degree * DEG2RAD)

    def tan(self) -> float:
        """calculate tarngetn

        Returns:
            float: tarngetn value
        """
        return math.tan(self._degree * DEG2RAD)

    @staticmethod
    def rad2deg(rad: float) -> float:
        """static utility. convert radian to degree

        Args:
            rad (float): radian value

        Returns:
            float: degree value
        """
        return rad * RAD2DEG

    @staticmethod
    def deg2rad(deg: float) -> float:
        """static utility. convert degree to radian

        Args:
            deg (float): degree value

        Returns:
            float: radian value
        """
        return deg * DEG2RAD

    @staticmethod
    def cos_deg(deg: float) -> float:
        """static utility. calculate cosine value for degree angle

        Args:
            deg (float): degree value

        Returns:
            float: cosine value
        """
        return math.cos(AngleDeg.deg2rad(deg))

    @staticmethod
    def sin_deg(deg: float) -> float:
        """static utility. calculate sine value for degree angle

        Args:
            deg (float): degree value

        Returns:
            float: sine value
        """
        return math.sin(AngleDeg.deg2rad(deg))

    @staticmethod
    def tan_deg(deg: float) -> float:
        """static utility. calculate tangent value for degree angle

        Args:
            deg (float): degree value

        Returns:
            float: tangent value
        """
        return math.tan(AngleDeg.deg2rad(deg))

    @staticmethod
    def acos_deg(cosine: float) -> float:
        """static utility. calculate arc cosine value

        Args:
            cosine (float): cosine value

        Returns:
            float: arc cosine value, that is degree type.
        """
        if cosine >= 1.0:
            return 0.0
        elif cosine <= -1.0:
            return 180.0
        else:
            return AngleDeg.rad2deg(math.acos(cosine))

    @staticmethod
    def asin_deg(sine: float) -> float:
        """static utility. calculate arc sine value

        Args:
            sine (float): sine value

        Returns:
            float: arc sine value, that is degree type.
        """
        if sine >= 1.0:
            return 90.0
        elif sine <= -1.0:
            return -90.0
        else:
            return AngleDeg.rad2deg(math.asin(sine))

    @staticmethod
    def atan_deg(tangent: float) -> float:
        """static utility. calculate arc tangent value

        Args:
            tangent (float): tangent value

        Returns:
            float: arc tangent value, that is degree.
        """
        return AngleDeg.rad2deg(math.atan(tangent))

    @staticmethod
    def atan2_deg(y_val: float, x_val: float) -> float:
        """static utility. calculate arc tangent value from XY

        Args:
            y_val (float): coordinate Y
            x_val (float): coordinate X

        Returns:
            float: arc tangent value.
        """
        if math.fabs(x_val) < EPSILON and math.fabs(y_val) < EPSILON:
            return 0.0
        else:
            return AngleDeg.rad2deg(math.atan2(y_val, x_val))

    @staticmethod
    def bisect(left: Union[int, float, AngleDeg],
        right: Union[int, float, AngleDeg]) -> Union[AngleDeg, float]:
        """static utility that returns bisect angle of [left, right]
        this method can take obtuse angle

        Args:
            left (Union[int, float, AngleDeg]): left start angle
            right (Union[int, float, AngleDeg]): right end angle

        Returns:
            Union[AngleDeg, float]: bisect angle
        """
        result = AngleDeg(left)
        rel = AngleDeg(AngleDeg(right) - result)
        diff = result.degree() - AngleDeg(right).degree()
        half_deg = rel.degree() * 0.5
        result += half_deg
        if (0.0 < diff < 180.0) or diff < -180.0:
            return result
        return result + 180.0

    @staticmethod
    def normalize_angle(dir: float):
        if dir < -360 or dir > 360:
            dir = dir % 360
        if dir < -180:
            dir += 360
        if dir > 180:
            dir -= 360
        return dir


""" test_polygon_2d.py file
    to test pyrusgeom Polygon2D class
"""
from cmath import sqrt
from unittest import TestCase
from pyrusgeom.polygon_2d import Polygon2D
from pyrusgeom.vector_2d import Vector2D
from pyrusgeom.rect_2d import Rect2D

class TestPolygon2D(TestCase):
    """Polygon2D class

    Args:
        TestCase (UnitTest): fail if any of tests falis
    """
    input_points_0 = [Vector2D(7, 7), Vector2D(7, -7), Vector2D(-7, -7), Vector2D(-7, 7),
                      Vector2D(9, 0), Vector2D(-9, 0), Vector2D(0, 9), Vector2D(0, -9)]
    input_points_1 = [Vector2D(0,0),Vector2D(0,3),Vector2D(4,0)]
    input_points_2 = [Vector2D(10, 10), Vector2D(0, 15) , Vector2D(-10, 10),
                      Vector2D(-10, -10), Vector2D(10, -10)]
    input_points_3 = [Vector2D(10, 10), Vector2D(-10, 10), Vector2D(-10, -10), Vector2D(10, -10)]

    def test_polygon_init_assign(self):
        plg_0 = Polygon2D()
        plg_1 = Polygon2D(self.input_points_0)
        plg_2 = Polygon2D()
        plg_2.assign(self.input_points_0)
        self.assertCountEqual(plg_0.vertices(),[Vector2D(0,0)])
        self.assertCountEqual(plg_1.vertices(),self.input_points_0)
        self.assertCountEqual(plg_2.vertices(),self.input_points_0)

    def test_add_vertex(self):
        plg_1 = Polygon2D(self.input_points_3)
        plg_1.add_vertex(Vector2D(0,15))
        self.assertCountEqual(plg_1.vertices(),self.input_points_2)

    def test_get_bounding_box(self):
        plg_0 = Polygon2D(self.input_points_0)
        rect_0 = plg_0.get_bounding_box()
        rect_ans = Rect2D(Vector2D(-9,-9),18,18)
        self.assertEqual(rect_0.top_right(),rect_ans.top_right())
        self.assertEqual(rect_0.bottom_left(),rect_ans.bottom_left())

    def test_bounding_box_center(self):
        plg_0 = Polygon2D(self.input_points_0)
        rect_0 = plg_0.bounding_box_center()
        rect_ans = Rect2D(Vector2D(-9,-9),18,18)
        self.assertEqual(rect_0,rect_ans.center())

    def test_contains(self):
        plg_0 = Polygon2D(self.input_points_2)
        alpha_point = Vector2D(14,-11)
        beta_point = Vector2D(0,0)
        gamma_point = Vector2D(5,14)
        delta_point = Vector2D(-10,10)
        self.assertFalse(plg_0.contains(alpha_point))
        self.assertTrue(plg_0.contains(beta_point))
        self.assertFalse(plg_0.contains(gamma_point))
        self.assertTrue(plg_0.contains(delta_point))
        self.assertFalse(plg_0.contains(delta_point,False))

    def test_dist(self):
        plg_0 = Polygon2D(self.input_points_2)
        alpha_point = Vector2D(0,-15)
        beta_point = Vector2D(0,0)
        gamma_point = Vector2D(-11,11)
        delta_point = Vector2D(10,15)
        self.assertEqual(plg_0.dist(alpha_point),5)
        self.assertEqual(plg_0.dist(beta_point),0)
        self.assertEqual(plg_0.dist(beta_point,False),10)
        self.assertEqual(plg_0.dist(gamma_point),sqrt(2))
        self.assertEqual(plg_0.dist(delta_point),50/sqrt(125))

    def test_double_signed_area(self):
        plg_0 = Polygon2D()
        plg_1 = Polygon2D(self.input_points_0)
        plg_2 = Polygon2D(self.input_points_1)
        plg_3 = Polygon2D(self.input_points_2)
        plg_4 = Polygon2D(self.input_points_3)

        self.assertEqual(plg_0.double_signed_area(),0)
        self.assertEqual(plg_1.double_signed_area(),-375)
        self.assertEqual(plg_2.double_signed_area(),-12)
        self.assertEqual(plg_3.double_signed_area(),900)
        self.assertEqual(plg_4.double_signed_area(),800)

    def test_get_rectangle_clipped_polygon(self):
        plg_3 = Polygon2D(self.input_points_2)
        ans_plg = [Vector2D(10, -10), Vector2D(10.0, -20.0),
                   Vector2D(-10.0, -20.0), Vector2D(-10, -10)]
        rect = Rect2D(Vector2D(-10, -20), 20, 20)
        plg_5 = plg_3.get_rectangle_clipped_polygon(rect)
        self.assertEqual(plg_5.vertices(),ans_plg)



import unittest
from pyrusgeom.vector_2d import Vector2D
from pyrusgeom.angle_deg import AngleDeg


class Vector2DTest(unittest.TestCase):
    def test_general_test(self):
        a = Vector2D(1, 2)
        self.assertEqual(a.x(), 1)

    def test_constractor(self):
        a = Vector2D(1.1, 2.1)
        self.assertEqual(a.x(), 1.1)
        b = Vector2D(1, 1)
        c = Vector2D(b)
        self.assertEqual(c.x(), 1)

    def test_assign(self):
        a = Vector2D(1.1, 2.1)
        a.assign(1, 2)
        self.assertEqual(a.x(), 1)
        self.assertEqual(a.y(), 2)

    def test_set_functions(self):
        a = Vector2D(0, 0)
        a.set_x(1)
        self.assertEqual(a.x(), 1)
        a.set_y(2)
        self.assertEqual(a.y(), 2)
        a.set_x_y(3, 5)
        self.assertEqual(a.x(), 3)
        self.assertEqual(a.y(), 5)

        b = Vector2D(0, 0)
        b.set_polar(10, 45)
        self.assertTrue(7 <= b.x() <= 7.1)
        self.assertTrue(7 <= b.y() <= 7.1)
        b.set_polar(10, AngleDeg(45))
        self.assertTrue(7 <= b.x() <= 7.1)
        self.assertTrue(7 <= b.y() <= 7.1)
        b.set_polar(10, -45)
        self.assertTrue(7 <= b.x() <= 7.1)
        self.assertTrue(-7.1 <= b.y() <= -7)
        b.set_polar(10, AngleDeg(-135))
        self.assertTrue(-7.1 <= b.x() <= -7)
        self.assertTrue(-7.1 <= b.y() <= -7)
        b.set_polar(10, AngleDeg(225))
        self.assertTrue(-7.1 <= b.x() <= -7)
        self.assertTrue(-7.1 <= b.y() <= -7)
        b.set_polar(10, 225)
        self.assertTrue(-7.1 <= b.x() <= -7)
        self.assertTrue(-7.1 <= b.y() <= -7)
        b.set_polar(10, 360)
        self.assertTrue(9.9 <= b.x() <= 10.1)
        self.assertTrue(-0.1 <= b.y() <= 0.1)
        a = Vector2D(4, 4)
        a.set_dir(0)
        self.assertEqual(a.th().degree(), 0)
        self.assertTrue(5.6 < a.x() < 5.7)

    def test_th(self):
        a = Vector2D(10, 10)
        b = a.th()
        self.assertIsInstance(b, AngleDeg)
        c = b.degree()
        self.assertTrue(44 <= c <= 46)

    def test_add(self):
        a = Vector2D(0, 0)
        b = Vector2D(2, 4)
        a.add(b)
        self.assertEqual(a.x(), 2)
        self.assertEqual(a.y(), 4)
        a.add_x(1)
        a.add_y(2)
        self.assertEqual(a.x(), 3)
        self.assertEqual(a.y(), 6)

    def test_scale(self):
        a = Vector2D(2, 3)
        a.scale(2)
        self.assertEqual(a.x(), 4)
        self.assertEqual(a.y(), 6)

    def test_copy(self):
        a = Vector2D(0, 0)
        b = a.copy()
        b.set_x(10)
        self.assertFalse(a.x() == 10)

    def test_dist(self):
        a = Vector2D(10, 0)
        b = Vector2D(20, 0)
        self.assertEqual(a.dist(b), 10)

    def test_length(self):
        a = Vector2D(2, 0)
        a.set_length(8)
        self.assertEqual(a.x(), 8)
        self.assertEqual(a.y(), 0)
        a.normalize()
        self.assertEqual(a.x(), 1)
        self.assertEqual(a.y(), 0)

    def test_rotate(self):
        a = Vector2D(10, 0)
        a.rotate(90)
        self.assertEqual(a.y(), 10)
        self.assertTrue(-0.01 < a.x() < 0.01)
        a = Vector2D(10, 0)
        a.rotate(AngleDeg(90))
        self.assertEqual(a.y(), 10)
        self.assertTrue(-0.01 < a.x() < 0.01)

    def test_operation(self):
        a = Vector2D(1, 1)
        b = Vector2D(4, 4)
        c = a + b
        self.assertEqual(c.x(), 5)
        self.assertEqual(c.y(), 5)
        c = b - a
        self.assertEqual(c.x(), 3)
        self.assertEqual(c.y(), 3)
        c = b / 2
        self.assertEqual(c.x(), 2)
        self.assertEqual(c.y(), 2)
        c = b * 2
        self.assertEqual(c.x(), 8)
        self.assertEqual(c.y(), 8)
        b = Vector2D(4, 4)
        b *= 2
        self.assertEqual(b.x(), 8)
        self.assertEqual(b.y(), 8)
        self.assertEqual(c.x(), 8)
        self.assertEqual(c.y(), 8)
        b = Vector2D(4, 4)
        b /= 2
        self.assertEqual(b.x(), 2)
        self.assertEqual(b.y(), 2)
        b = Vector2D(4, 4)
        b += Vector2D(1, 1)
        self.assertEqual(b.x(), 5)
        self.assertEqual(b.y(), 5)
        b = Vector2D(4, 4)
        b -= Vector2D(1, 1)
        self.assertEqual(b.x(), 3)
        self.assertEqual(b.y(), 3)

    def test_static(self):
        a = Vector2D.polar2vector(3, 0)
        self.assertEqual(a.x(), 3)
        self.assertEqual(a.y(), 0)
        a = Vector2D.from_polar(3, 0)
        self.assertEqual(a.x(), 3)
        self.assertEqual(a.y(), 0)


    def test_kwargs(self):
        a = Vector2D(y=1, x=2)
        self.assertEqual(a.x(), 2)
        self.assertEqual(a.y(), 1)

    def test_kwargs_polar(self):
        a = Vector2D(r=3, a=0)
        self.assertEqual(a.x(), 3)
        self.assertEqual(a.y(), 0)
        self.assertEqual(a.th(), 0)
        self.assertEqual(a.r(), 3)
        a = Vector2D(r=10, a=45)
        self.assertEqual(a.th(), 45)
        self.assertEqual(a.r(), 10)

if __name__ == '__main__':
    unittest.main()



""" segment_2d.py file
    Segment2D: class name
    Class attributes: _origin,_terminal
    TODO: fix the circular import tri-segmant

"""
from __future__ import annotations
from typing import Union
import math

from pyrusgeom.triangle_2d import Triangle2D
from pyrusgeom.line_2d import Line2D
from pyrusgeom.vector_2d import Vector2D
from pyrusgeom.angle_deg import AngleDeg
from pyrusgeom.math_values import CALC_ERROR, EPSILON


class Segment2D:
    """ Handling segments in SS2D

    Attributes:
        _origin: a Vector2D for orgin point
        _terminal: a Vector2D for terminal point
    """

    def __init__(self, *args) -> None:
        """This is the class init function for Segment2D.

        Defualt:
            create a segment starts at (0,0) and ends at (0,0)
        OR:
            bulid a segment2d with given values

        Args:
            four:
                4 float:
                    float: origin_x 1st point x value of segment edge
                    float: origin_y 1st point y value of segment edge
                    float: terminal_x 2nd point x value of segment edge
                    float: terminal_y 2nd point y value of segment edge
            three:
                1 vector 2 float:
                    Vector2D: origin point
                    float: length of line segment
                    float: line direction from origin point
            two:
                2 vector:
                    Vector2D: origin 1st point of segment edge
                    Vector2D: terminal 2nd point of segment edge
            none:
                Default

        Raises:
            Exception: input must be (4 float) or (1 vector2d , 2 float) or (2 vector2d) or nothing
        """
        self._origin: Vector2D = Vector2D(0, 0)
        self._terminal: Vector2D = Vector2D(0, 0)
        if len(args) == 2:
            self._origin = Vector2D(args[0])
            self._terminal = Vector2D(args[1])
        elif len(args) == 3:
            self._origin = Vector2D(args[0])
            self._terminal = args[0] + Vector2D.from_polar(args[1], args[2])
        elif len(args) == 4:
            self._origin = Vector2D(args[0], args[1])
            self._terminal = Vector2D(args[2], args[3])
        elif len(args) == 0:
            self._origin: Vector2D = Vector2D(0, 0)
            self._terminal: Vector2D = Vector2D(0, 0)
        else:
            raise Exception("input must be (4 float) or \
                (1 vector2d , 2 float) or (2 vector2d) or nothing")

    def assign(self, *args) -> None:
        """same as __init__

        Defualt:
            reset the segment starts at (0,0) and ends at (0,0)
        OR:
            refill the segment2d with given values

        Args:
            four:
                4 float:
                    float: origin_x 1st point x value of segment edge
                    float: origin_y 1st point y value of segment edge
                    float: terminal_x 2nd point x value of segment edge
                    float: terminal_y 2nd point y value of segment edge
            three:
                1 vector 2 float:
                    Vector2D: origin point
                    float: length of line segment
                    float: line direction from origin point
            two:
                2 vector:
                    Vector2D: origin 1st point of segment edge
                    Vector2D: terminal 2nd point of segment edge
            none:
                Default

        Raises:
            Exception: input must be (4 float) or (1 vector2d , 2 float) or (2 vector2d) or nothing
        """
        # self.__init__(args)
        self._origin: Vector2D = Vector2D(0, 0)
        self._terminal: Vector2D = Vector2D(0, 0)
        if len(args) == 2:
            self._origin = Vector2D(args[0])
            self._terminal = Vector2D(args[1])
        elif len(args) == 3:
            self._origin = Vector2D(args[0])
            self._terminal = args[0] + Vector2D.from_polar(args[1], args[2])
        elif len(args) == 4:
            self._origin = Vector2D(args[0], args[1])
            self._terminal = Vector2D(args[2], args[3])
        elif len(args) == 0:
            self._origin: Vector2D = Vector2D(0, 0)
            self._terminal: Vector2D = Vector2D(0, 0)
        else:
            raise Exception("input must be (4 float) or \
                (1 vector2d , 2 float) or (2 vector2d) or nothing")

    def is_valid(self) -> bool:
        """check if self line segment is valid or not.

        origin's coordinates value have to be different from terminal's one.

        Returns:
            bool: True if origin is not equal with terminal. else False.
        """
        return not self._origin.equals_weakly(self._terminal)

    def origin(self) -> Vector2D:
        """get a copy of 1st point segment edge

        Returns:
            Vector2D: a Vector1d with the 1st point segment values
        """

        return Vector2D(self._origin)

    def origin_(self) -> Vector2D:
        """get 1st point of segment edge

        Returns:
            Vector2D: reference to the vector object
        """
        return self._origin

    def terminal(self) -> Vector2D:
        """get a copy of 2nd point segment edge

        Returns:
            Vector2D: a Vector1d with the 2nd point segment values
        """

        return Vector2D(self._terminal)

    def terminal_(self) -> Vector2D:
        """get og 2nd point of segment edge

        Returns:
            Vector2D: reference to the vector object
        """
        return self._terminal

    def line(self) -> Line2D:
        """get line generated from segment

        Returns:
            Line2D: generated line object
        """
        return Line2D(self._origin, self._terminal)

    def length(self) -> float:
        """get the length of this segment

        Returns:
            float: distance value
        """
        return self._origin.dist(self._terminal)

    def direction(self) -> AngleDeg:
        """get the direction angle of self line segment

        Returns:
            AngleDeg: angle object
        """
        return (self._terminal - self._origin).th()

    def swap(self) -> Segment2D:
        """swaps segment edge point and returns it

        Returns:
            Segment2D: self object
        """

        tmp = self._origin
        self._origin = self._terminal
        self._terminal = tmp
        return self

    def copy(self) -> Segment2D:
        """get a copy of this Segment

        Returns:
            Segment2D: copy object
        """
        return Segment2D(self.origin(), self.terminal())

    def reverse(self) -> Segment2D:
        """swap segment edge point and returns it.

        This method is equivalent to swap()

        Returns:
            Segment2D: self object
        """
        return self.swap()

    def reversed_segment(self) -> Segment2D:
        """create a copy reversed line segment.

        Returns:
            Segment2D: reversed copy object
        """
        return Segment2D(self._origin, self._terminal).reverse()

    def perpendicular_bisector(self) -> Line2D:
        """make perpendicular bisector line from segment points

        Returns:
            Line2D: line object
        """
        return Line2D.perpendicular_bisector(self._origin, self._terminal)

    def contains(self, point: Vector2D) -> bool:
        """check if the point is within the CALC_ERROR rectangle
            defined by this segment as a diagonal line.
        Args:
            point (Vector2D): considered point

        Returns:
            bool: True if rectangle contains point. else False
        """
        return (point.x() - self._origin.x()) * (point.x() - self._terminal.x()) <= CALC_ERROR and (
        point.y() - self._origin.y()) * (point.y() - self._terminal.y()) <= CALC_ERROR

    def __eq__(self, other: Segment2D) -> bool:
        """check if self line segment has completely same value as input line segment.

        Args:
            other (Segment2D): compared object.

        Returns:
            bool: true if same. false if not.
        """
        return self._origin == other.origin_() and self._terminal == other.terminal_()

    def equals_weakly(self, other: Segment2D) -> bool:
        """check if self line segment has weakly same value as input line segment.

        Args:
            other (Segment2D): compared object.

        Returns:
            bool: true if almostequals. false if not.
        """
        return self._origin.equals_weakly(other.origin()) and \
            self._terminal.equals_weakly(other.terminal())

    def projection(self, point: Vector2D) -> Vector2D:
        """calculates projection point from input point

        if it does not exist, the invalidated value vector is returned.

        Args:
            point (Vector2D): input point

        Returns:
            Vector2D: projection point from input point.
        """
        direction = self._terminal - self._origin
        length = direction.r()

        if length < EPSILON:
            return self._origin

        direction /= length  # normalize

        direction_ip = direction.inner_product(point - self._origin)
        if -EPSILON < direction_ip < length + EPSILON:
            direction *= direction_ip
            tmp_vec = Vector2D(self._origin)
            tmp_vec += direction
            return tmp_vec

        return Vector2D.invalid()

    def intersection(self, *args) -> Vector2D:
        """check & get the intersection point with other line or segment

        Args:
            one:
                Line2D: checked line object
            two:
                Segment2D: checked segment
                Boolean: if it value is False, end point is disallowed as an intersection.

        if it does not exist, the invalidated value vector is returned.

        Returns:
            Vector2D: intersection point.
        """
        if isinstance(args[0], Line2D):
            tmp_line = self.line()
            sol = tmp_line.intersection(args[0])
            if not sol.is_valid() or not self.contains(sol):
                return Vector2D.invalid()
            return sol

        elif isinstance(args[0], Segment2D):
            allow_end_point = False
            if len(args) == 2 and isinstance(args[1], bool):
                allow_end_point = args[1]
            sol = self.line().intersection(args[0].line())
            if not sol.is_valid() or not self.contains(sol) or not args[0].contains(sol):
                return Vector2D.invalid()
            if not allow_end_point and not self.exist_intersection_except_endpoint(args[0]):
                return Vector2D.invalid()
            return sol

        else:
            return Vector2D.invalid()

    def exist_intersection(self, other: Union[Segment2D, Line2D]) -> bool:
        """check if segment and other(line/segment) cross each other or not.

        Args:
            other (Union[Segment2D,Line2D]): other object for cross checking

        Returns:
            bool: True if this segment crosses. else returns False
        """

        if isinstance(other, Segment2D):
            tri_a0 = Triangle2D.double_signed_area_st(
                self._origin, self._terminal, other.origin())
            tri_a1 = Triangle2D.double_signed_area_st(
                self._origin, self._terminal, other.terminal())
            tri_b0 = Triangle2D.double_signed_area_st(
                other.origin(), other.terminal(), self._origin)
            tri_b1 = Triangle2D.double_signed_area_st(
                other.origin(), other.terminal(), self._terminal)

            if tri_a0 * tri_a1 < 0.0 and tri_b0 * tri_b1 < 0.0:
                return True

            if self._origin == self._terminal:
                if other.origin() == other.terminal():
                    return self._origin == other.origin()

                return tri_b0 == 0.0 and other.check_intersects_on_line(self._origin)

            if other.origin() == other.terminal():
                return tri_a0 == 0.0 and self.check_intersects_on_line(other.origin())

            if tri_a0 == 0.0 and self.check_intersects_on_line(other.origin()) or (
                    tri_a1 == 0.0 and self.check_intersects_on_line(other.terminal())) or (
                    tri_b0 == 0.0 and other.check_intersects_on_line(self._origin)) or (
                    tri_b1 == 0.0 and other.check_intersects_on_line(self._terminal)):
                return True
            return False
        if isinstance(other, Line2D):
            tri_a0 = other.a() * self._origin.x() + other.b() * self._origin.y() + other.c()
            tri_a1 = other.a() * self._terminal.x() + other.b() * \
                self._terminal.y() + other.c()
            return tri_a0 * tri_a1 <= 0.0

    def check_intersects_on_line(self, point: Vector2D) -> bool:
        """check if given point intersects on line

        Args:
            point (Vector2D): input point

        Returns:
            bool: True if intersects. else False
        """
        if self._origin.x() == self._terminal.x():
            return (self._origin.y() <= point.y() <= self._terminal.y()) or (
                self._terminal.y() <= point.y() <= self._origin.y())
        return (self._origin.x() <= point.x() <= self._terminal.x()) or (
            self._terminal.x() <= point.x() <= self._origin.x())

    def exist_intersection_except_endpoint(self, other: Segment2D) -> bool:
        """check if segments intersect each other on non terminal point.

        Args:
            other (Segment2D): other segment

        Returns:
            bool: True if segments intersect and intersection point is not a
                terminal point of segment.
                False if segments do not intersect or intersect on terminal point of segment.
        """
        return (Triangle2D.double_signed_area_st(
            self._origin,
            self._terminal,
            other.origin())
            * Triangle2D.double_signed_area_st(
                self._origin,
            self._terminal,
            other.terminal()) < 0.0) and (
            Triangle2D.double_signed_area_st(
                other.origin(),
                other.terminal(),
                self._origin)
            * Triangle2D.double_signed_area_st(
                other.origin(),
                other.terminal(),
                self._terminal) < 0.0)

    def intersects_except_endpoint(self, other: Segment2D) -> bool:
        """check if segments intersect each other on non terminal point.

        This method is equivalent to exist_intersection_except_endpoint()

        Args:
            other (Segment2D): other segment

        Returns:
            bool: True if segments intersect and intersection point is not a
                terminal point of segment.
                False if segments do not intersect or intersect on terminal point of segment.
        """
        return self.exist_intersection_except_endpoint(other)

    def nearest_point(self, point: Vector2D) -> Vector2D:
        """find minimal distance point on segment from input point.

        Args:
            point (Vector2D): input point.

        Returns:
            Vector2D: nearest point on segment. if multiple nearest points found.
                    returns one of them.
        """
        vec_tmp = self._terminal - self._origin

        len_square = vec_tmp.r2()

        if len_square == 0.0:
            return self._origin

        inner_product = vec_tmp.inner_product((point - self._origin))

        if inner_product <= 0.0:
            return self._origin

        if inner_product >= len_square:
            return self._terminal

        return self._origin + vec_tmp * inner_product / len_square

    def dist(self, other: Union[Segment2D, Vector2D]) -> float:
        """get minimum distance between this segment and other input

        Args:
            other (Union[Segment2D, Vector2D]): point or another segment

        Raises:
            Exception: Input must be a Vector2D or a Segement2D

        Returns:
            float: minimum distance between these two
        """

        if isinstance(other, Vector2D):
            other: Vector2D = other
            length = self.length()
            if length == 0.0:
                return self._origin.dist(other)
            tmp_vec = self._terminal - self._origin
            prod = tmp_vec.inner_product(other - self._origin)
            if 0.0 <= prod <= length * length:
                return math.fabs(Triangle2D.double_signed_area_st(
                    self._origin,
                    self._terminal,
                    other) / length)
            return math.sqrt(min(self._origin.dist2(other),
                                 self._terminal.dist2(other)))

        if isinstance(other, Segment2D):
            other: Segment2D = other
            if self.exist_intersection(other):
                return 0.0
            return min(self.dist(other.origin()),
                       self.dist(other.terminal()),
                       other.dist(self._origin),
                       other.dist(self._terminal))

        raise Exception("Input must be a Vector2D or a Segement2D")

    def farthest_dist(self, point: Vector2D) -> float:
        """get maximum distance between this segment and input point

        Args:
            point (Vector2D): input point

        Returns:
            float: maximum distance between this segment and point
        """
        return math.sqrt(max(self._origin.dist2(point), self._terminal.dist2(point)))

    def on_segment(self, point: Vector2D) -> float:
        """strictly check if point is on segment or not

        Args:
            point (Vector2D): input point

        Returns:
            float: True if point is on this segment. else False.
        """
        return Triangle2D.double_signed_area_st(self._origin, self._terminal, point) == 0.0 \
            and self.check_intersects_on_line(point)

    def on_segment_weakly(self, point: Vector2D) -> bool:
        """weakly check if point is on segment or not

        Args:
            point (Vector2D): input point

        Returns:
            bool: True if point is on this segment. else False.
        """
        projection = self.projection(point)
        return projection.is_valid() and point.equals_weakly(projection)

    def __repr__(self) -> str:
        """represent Segment2D as a string

        Returns:
            str: Circle2D's _origin and _terminal as string
        """
        return f"[{self._origin},{self._terminal}]"

    def to_str(self, ostr: str) -> None:
        """add segment to the ostr

        Args:
            ostr (str): str to add to it
        """
        ostr += f' (line {round(self.origin().x(), 3)} {round(self.origin().y(), 3)} \
            {round(self.terminal().x(), 3)} {round(self.terminal().y(), 3)})'



""" sector_2d.py file
    Sector2D: class name
    Class attributes : _center,_min_r,_max_r,_start,_end,is_valid
    TODO: add test and reverse
"""
from __future__ import annotations
import math
from typing import Union

from pyrusgeom.region_2d import Region2D
from pyrusgeom.vector_2d import Vector2D
from pyrusgeom.angle_deg import AngleDeg
from pyrusgeom.math_values import EPSILON, PI


class Sector2D(Region2D):
    """handling Sectors in SS2D

    Args:
        Region2D ([type]): created from Region2D class

    Attributes:
        _center: center point
        _min_r: smaller radius
        _max_r: bigger radius
        _start: start angle(turn clockwise)
        _end: end angle(turn clockwise)
        is_valid: is it valid or not
    """

    def __init__(self, center: Vector2D, min_r: float, max_r: float,
                 start: Union[AngleDeg, float], end: Union[AngleDeg, float]) -> None:
        """This is the class init function for sector.

        constructor with all variables

        Args:
            center (Vector2D): center point
            min_r (float): smaller radius
            max_r (float): bigger radius
            start (Union[AngleDeg, float]): start angle(turn clockwise)
            end (Union[AngleDeg, float]): end angle(turn clockwise)
        """
        super().__init__()

        self._center = Vector2D(center)
        self._min_r = min(0.0, min_r)
        self._max_r = max(min_r, max_r)
        self._start = AngleDeg(start)
        self._end = AngleDeg(end)
        self.is_valid = True

    def assign(self, center: Vector2D, min_r: float, max_r: float,
               start: Union[AngleDeg, float], end: Union[AngleDeg, float]) -> None:
        """assign new value

        Args:
            center (Vector2D): center point
            min_r (float): smaller radius
            max_r (float): bigger radius
            start (Union[AngleDeg, float]): start angle(turn clockwise)
            end (Union[AngleDeg, float]): end angle(turn clockwise)
        """
        self._center = Vector2D(center)
        self._min_r = min(0.0, min_r)
        self._max_r = max(min_r, max_r)
        self._start = AngleDeg(start)
        self._end = AngleDeg(end)

    def center(self) -> Vector2D:
        """get the center point copy

        Returns:
            Vector2D: a Vector2D with center point values
        """
        return Vector2D(self._center)

    def center_(self) -> Vector2D:
        """get the center point

        Returns:
            Vector2D: reference to the og member variable
        """
        return self._center

    def radius_min(self) -> float:
        """get the small side radius

        Returns:
            float: the small side radius
        """
        return self._min_r

    def radius_max(self) -> float:
        """get the big side radius

        Returns:
            float: the big side radius
        """
        return self._min_r

    def angle_left_start(self) -> AngleDeg:
        """get the left start angle copy

        Returns:
            AngleDeg: an AngleDeg with start values
        """
        return AngleDeg(self._start)

    def angle_left_start_(self) -> AngleDeg:
        """get the left start angle

        Returns:
            AngleDeg: reference to the og member variable
        """
        return self._start

    def angle_right_end(self) -> AngleDeg:
        """get the right end angle copy

        Returns:
            AngleDeg: an AngleDeg with start values
        """
        return AngleDeg(self._end)

    def angle_right_end_(self) -> AngleDeg:
        """get the right end angle

        Returns:
            AngleDeg: reference to the og member variable
        """
        return self._end

    def area(self) -> float:
        """calculate the area of this sector

        Returns:
            float: the value of area
        """
        circle_area = self._max_r * self._max_r * PI - self._min_r * self._min_r * PI
        div = (self._end - self._start).degree_()
        if div < 0.0:
            div += 360.0
        return circle_area * div / 360.0

    def contains(self, point: Vector2D) -> bool:
        """check if point is within this sector

        Args:
            point (Vector2D): considered point

        Returns:
            bool: True if contains. else False.
        """
        rel = point - self._center
        delta = rel.r2()
        return (self._min_r * self._min_r <= delta <= self._max_r * self._max_r and
                rel.th().is_within(self._start, self._end))

    def get_circumference_min(self) -> float:
        """get smaller side circumference

        Returns:
            float: the length of circumference
        """
        div = (self._end - self._start).degree_()
        if div < 0.0:
            div += 360.0
        return (2.0 * self._min_r * PI) * (div / 360.0)

    def get_circumference_max(self) -> float:
        """get bigger side circumference

        Returns:
            float: the length of circumference
        """
        div = (self._end - self._start).degree_()
        if div < 0.0:
            div += 360.0
        return (2.0 * self._max_r * PI) * (div / 360.0)

    def __eq__(self, other: Sector2D) -> bool:
        """operator == for Sector2D

        Args:
            other (Sector2D): right hand side argument
        Returns:
            bool: true if equal or difference is less than EPSILON. else false
        """
        return (self._center.equals_weakly(other.center_()) and
            self._start == other.angle_left_start_() and
            self._end == other.angle_right_end_() and
            math.fabs(self._min_r - other.radius_min()) < EPSILON and
            math.fabs(self._max_r - other.radius_max()) < EPSILON)
        # return super().__eq__(other)

    def __repr__(self) -> str:
        """represent Sector2D as a string

        Returns:
            str: Sector2D's _center,_start,_end,_min_r and _max_r as string
        """
        return f'[v({self._center}) a({self._start}, {self._end}) \
         r({self._min_r}, {self._max_r})]'
